import itertools
from typing import Set

import networkx as nx
from networkx import DiGraph

from sqllineage.core import LineageResult
from sqllineage.models import Table


class CombinedLineageResult:
    def __init__(self, graph: DiGraph) -> None:
        """
        The combined lineage result in representation of Directed Acyclic Graph.

        :param graph: the Directed Acyclic Graph holding all the combined lineage result.
        """
        self._graph = graph
        self._selfloop_tables = self.__retrieve_tag_tables("selfloop")
        self._sourceonly_tables = self.__retrieve_tag_tables("source")
        self._targetonly_tables = self.__retrieve_tag_tables("target")

    @property
    def source_tables(self) -> Set[Table]:
        """
        a list of source :class:`sqllineage.models.Table`
        """
        source_tables = {
            table for table, deg in self._graph.in_degree if deg == 0
        }.intersection({table for table, deg in self._graph.out_degree if deg > 0})
        source_tables |= self._selfloop_tables
        source_tables |= self._sourceonly_tables
        return source_tables

    @property
    def target_tables(self) -> Set[Table]:
        """
        a list of target :class:`sqllineage.models.Table`
        """
        target_tables = {
            table for table, deg in self._graph.out_degree if deg == 0
        }.intersection({table for table, deg in self._graph.in_degree if deg > 0})
        target_tables |= self._selfloop_tables
        target_tables |= self._targetonly_tables
        return target_tables

    @property
    def intermediate_tables(self) -> Set[Table]:
        """
        a list of intermediate :class:`sqllineage.models.Table`
        """
        intermediate_tables = {
            table for table, deg in self._graph.in_degree if deg > 0
        }.intersection({table for table, deg in self._graph.out_degree if deg > 0})
        intermediate_tables -= self.__retrieve_tag_tables("selfloop")
        return intermediate_tables

    @property
    def lineage_graph(self) -> DiGraph:
        """
        The DiGraph held by CombinedLineageResult
        """
        return self._graph

    def __retrieve_tag_tables(self, tag) -> Set[Table]:
        return {
            table
            for table, attr in self._graph.nodes(data=True)
            if attr.get("tag") == tag
        }


def combine(*args: LineageResult) -> CombinedLineageResult:
    """
    To combine multiple :class:`sqllineage.core.LineageResult` into :class:`sqllineage.combiners.CombinedLineageResult`
    """
    g = DiGraph()
    for lineage_result in args:
        if lineage_result.drop:
            for table in lineage_result.drop:
                if g.has_node(table) and g.degree[table] == 0:
                    g.remove_node(table)
        elif lineage_result.rename:
            for (table_old, table_new) in lineage_result.rename:
                g = nx.relabel_nodes(g, {table_old: table_new})
        else:
            read, write = lineage_result.read.copy(), lineage_result.write.copy()
            if lineage_result.intermediate:
                read -= lineage_result.intermediate
            if len(read) > 0 and len(write) == 0:
                g.add_nodes_from(read, tag="source")
            elif len(read) == 0 and len(write) > 0:
                g.add_nodes_from(write, tag="target")
            else:
                g.add_nodes_from(read)
                g.add_nodes_from(write)
                for source, target in itertools.product(read, write):
                    g.add_edge(source, target)
    for table in {e[0] for e in nx.selfloop_edges(g)}:
        g.nodes[table]["tag"] = "selfloop"
    return CombinedLineageResult(g)



import logging
from typing import Dict, List

import sqlparse
from sqlparse.sql import Statement

from sqllineage.combiners import combine
from sqllineage.core import LineageAnalyzer
from sqllineage.drawing import draw_lineage_graph
from sqllineage.io import to_cytoscape
from sqllineage.models import Table

logger = logging.getLogger(__name__)


def lazy_method(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        if not self._evaluated:
            self._eval()
        return func(*args, **kwargs)

    return wrapper


def lazy_property(func):
    return property(lazy_method(func))


class LineageRunner(object):
    def __init__(
        self,
        sql: str,
        encoding: str = None,
        verbose: bool = False,
        draw_options: Dict[str, str] = None,
    ):
        """
        The entry point of SQLLineage after command line options are parsed.

        :param sql: a string representation of SQL statements.
        :param encoding: the encoding for sql string
        :param verbose: verbose flag indicate whether statement-wise lineage result will be shown
        """
        self._encoding = encoding
        self._sql = sql
        self._verbose = verbose
        self._draw_options = draw_options if draw_options else {}
        self._evaluated = False
        self._stmt: List[Statement] = []

    @lazy_method
    def __str__(self):
        """
        print out the Lineage Summary.
        """
        statements = self.statements(strip_comments=True)
        source_tables = "\n    ".join(str(t) for t in self.source_tables)
        target_tables = "\n    ".join(str(t) for t in self.target_tables)
        combined = f"""Statements(#): {len(statements)}
Source Tables:
    {source_tables}
Target Tables:
    {target_tables}
"""
        if self.intermediate_tables:
            intermediate_tables = "\n    ".join(
                str(t) for t in self.intermediate_tables
            )
            combined += f"""Intermediate Tables:
    {intermediate_tables}"""
        if self._verbose:
            result = ""
            for i, lineage_result in enumerate(self._lineage_results):
                stmt_short = statements[i].replace("\n", "")
                if len(stmt_short) > 50:
                    stmt_short = stmt_short[:50] + "..."
                content = str(lineage_result).replace("\n", "\n    ")
                result += f"""Statement #{i + 1}: {stmt_short}
    {content}
"""
            combined = result + "==========\nSummary:\n" + combined
        return combined

    @lazy_method
    def to_cytoscape(self) -> List[Dict[str, Dict[str, str]]]:
        """
        to turn the DAG into cytoscape format.
        """
        return to_cytoscape(self._combined_lineage_result.lineage_graph)

    def draw(self) -> None:
        """
        to draw the lineage directed graph
        """
        draw_options = self._draw_options
        if draw_options.get("f") is None:
            draw_options.pop("f", None)
            draw_options["e"] = self._sql
        return draw_lineage_graph(**draw_options)

    @lazy_method
    def statements(self, **kwargs) -> List[str]:
        """
        a list of statements.

        :param kwargs: the key arguments that will be passed to `sqlparse.format`
        """
        return [sqlparse.format(s.value, **kwargs) for s in self.statements_parsed]

    @lazy_property
    def statements_parsed(self) -> List[Statement]:
        """
        a list of :class:`sqlparse.sql.Statement`
        """
        return self._stmt

    @lazy_property
    def source_tables(self) -> List[Table]:
        """
        a list of source :class:`sqllineage.models.Table`
        """
        return sorted(self._combined_lineage_result.source_tables, key=lambda x: str(x))

    @lazy_property
    def target_tables(self) -> List[Table]:
        """
        a list of target :class:`sqllineage.models.Table`
        """
        return sorted(self._combined_lineage_result.target_tables, key=lambda x: str(x))

    @lazy_property
    def intermediate_tables(self) -> List[Table]:
        """
        a list of intermediate :class:`sqllineage.models.Table`
        """
        return sorted(
            self._combined_lineage_result.intermediate_tables, key=lambda x: str(x)
        )

    def _eval(self):
        self._stmt = [
            s
            for s in sqlparse.parse(self._sql.strip(), self._encoding)
            if s.token_first(skip_cm=True)
        ]
        self._lineage_results = [LineageAnalyzer().analyze(stmt) for stmt in self._stmt]
        self._combined_lineage_result = combine(*self._lineage_results)
        self._evaluated = True



import logging
import os
import sys
from argparse import Namespace
from urllib.parse import urlencode

from flask import Flask, jsonify, request
from flask_cors import CORS

from sqllineage import DEFAULT_PORT
from sqllineage import STATIC_FOLDRE
from sqllineage.helpers import extract_sql_from_args

logger = logging.getLogger(__name__)

app = Flask(
    __name__,
    static_url_path="",
    static_folder=os.path.join(os.path.dirname(__file__), STATIC_FOLDRE),
)
CORS(app)


@app.route("/")
def index():
    return app.send_static_file("index.html")


@app.route("/lineage", methods=["POST"])
def lineage():
    # this is to avoid circular import
    from sqllineage.runner import LineageRunner

    req_args = Namespace(**request.get_json())
    sql = extract_sql_from_args(req_args)
    resp = LineageRunner(sql).to_cytoscape()
    return jsonify(resp)


cli = sys.modules["flask.cli"]
cli.show_server_banner = lambda *x: None  # type: ignore


def draw_lineage_graph(**kwargs) -> None:
    port = kwargs.pop("p", DEFAULT_PORT)
    querystring = urlencode({k: v for k, v in kwargs.items() if v})
    print(f" * SQLLineage Running on http://localhost:{port}/?{querystring}")
    app.run(port=port)
